using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using AssetBundles;
using Battlehub.RTCommon;
using SimpleJSON;
using Stopper;
using UnityEngine;
using UnityEngine.UI;
using MVR.FileManagementSecure;

/// <summary>

/// </summary>

namespace Stopper {
	public class CUAControllers : MVRScript
	{
        string savePath = "Saves\\CUAController";
		string fileExt = "cuacontroller";


		// Prefabricated UI for joint controllers. Not sure how to get ahold of this
		// other than copying from somewhere
		GameObject controllerUIPrefab;
		
		// A sample controller for retrieving controller meshes. Again, not sure
		// how to find this other than copying from elsewhere
		FreeControllerV3 sampleController;
		
		// GameObjects to organize all of our stuff
		GameObject unityAssetControls;
		GameObject controlUI;
		GameObject controllers;

		Atom controlAtom;
		Transform cuaRoot;

		List<FreeControllerV3> createdControllers = new List<FreeControllerV3>();
		List<ConfigurableJoint> controllerJoints = new List<ConfigurableJoint>();

		List<ConfigurableJoint> createdJoints = new List<ConfigurableJoint>();
		List<Rigidbody> createdRigidbodies = new List<Rigidbody>();
		
		FreeControllerV3[] originalControllers;
		
		List<UnityEngine.Object> createdObjects = new List<UnityEngine.Object>();

		// UI
		string selectedRootPath;

		JSONStorableBool createRigidBodies;
		JSONStorableBool createJoints;
		JSONStorableBool loadOnStartup;

		bool loaded = false;

		JSONClass savedJson;

		float defaultRbMass = 1f;
		float defaultRbDrag = 1f;
		float defaultJointDriveSpring = 100f;
		float defaultJointDriveDamper = 0f;
		float defaultJointDriveMaxForce = 100f;
		
		public override void Init()
		{
			GameObject windowDisplayControl = GameObject.Find("/SceneAtoms/WindowCamera/reParentObject/DisplayControl");
			if(windowDisplayControl == null){
				SuperController.LogError("Unable to find window display control.");
				return;
			}

			sampleController = windowDisplayControl.GetComponent<FreeControllerV3>();
			
			//foreach(var s in containingAtom.GetStorableIDs()) {
			//	SuperController.LogError($"{s} {GetPathToRoot(containingAtom.GetStorableByID(s).transform, containingAtom.transform)}");
			//}

			JSONStorableStringChooser rootChooser = new JSONStorableStringChooser(
				"Root Node", 
				GetRootCandidates(),
				null, 
				"Root Node", 
				(string path) => {
					selectedRootPath = path;
				});
			CreateScrollablePopup(rootChooser, false);
			RegisterStringChooser(rootChooser);
			

			createRigidBodies = new JSONStorableBool("Create Rigidbodies", true);
			createJoints = new JSONStorableBool("Create Joints", true);
	
			var createRigidBodiesToggle = CreateToggle(createRigidBodies, false);
			RegisterBool(createRigidBodies);
			
			var createJointsToggle = CreateToggle(createJoints, false);
			RegisterBool(createJoints);

			loadOnStartup = new JSONStorableBool("LoadOnStartup", false);
			RegisterBool(loadOnStartup);
			
			UIDynamicButton goButton = CreateButton("Create Controllers", false);
			goButton.button.onClick.AddListener(() => {
				StartCoroutine("AddControllers");
			});

			var loadPresetUI = CreateButton("Load Preset");
			loadPresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var shortcuts = FileManagerSecure.GetShortCutsForDirectory(savePath);
				SuperController.singleton.GetMediaPathDialog(LoadPreset, fileExt, savePath, false, true, false, null, false, shortcuts);
			});

			var savePresetUI = CreateButton("Save Preset");
			savePresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var fileBrowserUI = SuperController.singleton.fileBrowserUI;
				fileBrowserUI.SetTitle("Save Preset");
				fileBrowserUI.fileRemovePrefix = null;
				fileBrowserUI.hideExtension = false;
				fileBrowserUI.keepOpen = false;
				fileBrowserUI.fileFormat = fileExt;
				fileBrowserUI.defaultPath = savePath;
				fileBrowserUI.showDirs = true;
				fileBrowserUI.shortCuts = null;
				fileBrowserUI.browseVarFilesAsDirectories = false;
				fileBrowserUI.SetTextEntry(true);
				fileBrowserUI.Show(SavePreset);
				fileBrowserUI.fileEntryField.text = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds + "." + fileExt;
				fileBrowserUI.ActivateFileNameField();
			});

			AddSlider("Rigidbody Drag", defaultRbDrag, 0f, 2f, (float val) => {
				foreach(var rb in createdRigidbodies) {
					rb.drag = val;
				}
			}, true, true);

			AddSlider("Rigidbody Mass", defaultRbMass, 0f, 2f, (float val) => {
				foreach(var rb in createdRigidbodies) {
					rb.mass = val;
				}
			}, true, true);

			AddSlider("Joint Spring", defaultJointDriveSpring, 0f, 1000f, (float val) => {
				foreach(var joint in createdJoints) {
					var drive = joint.slerpDrive;
					drive.positionSpring = val;
					joint.slerpDrive = drive;
				}
			}, true, true);

			AddSlider("Joint Damper", defaultJointDriveDamper, 0f, 10f, (float val) => {
				foreach(var joint in createdJoints) {
					var drive = joint.slerpDrive;
					drive.positionDamper = val;
					joint.slerpDrive = drive;
				}
			}, true, true);

			AddSlider("Joint Max Force", defaultJointDriveMaxForce, 0f, 1000f, (float val) => {
				foreach(var joint in createdJoints) {
					var drive = joint.slerpDrive;
					drive.maximumForce = val;
					joint.slerpDrive = drive;
				}
			}, true, true);

		}

		private void Start(){
			if(loadOnStartup.val) {
				StartCoroutine("AddControllers");
			}
		}

		private List<String> GetRootCandidates(){
			List<string> candidates = new List<string>();

			Queue<Transform> frontier = new Queue<Transform>();

			var asset = containingAtom.GetStorableByID("asset");
			if(asset != null) {
				cuaRoot = asset.transform;
			}
			if(cuaRoot == null) {
				cuaRoot = containingAtom.reParentObject.Find("object/rescaleObject");
			}
			if(cuaRoot == null) {
				cuaRoot = containingAtom.reParentObject;
			}

			frontier.Enqueue(cuaRoot);

			Transform node;
			while(frontier.Count > 0){
				node = frontier.Dequeue();
				if(frontier == null)
					break;
				for(int i = 0; i < node.childCount; i++){
					var child = node.GetChild(i);
					if(child.childCount == 0) {
						continue;
					}
					if(candidates.Count < 30){
						candidates.Add(GetPathToRoot(child, cuaRoot));
						frontier.Enqueue(child);
					} else {
						return candidates;
					}
				}
			}

			return candidates;
		}

		private string GetPathToRoot(Transform t, Transform root){
			if (t.parent == null || t.parent == root ){
				return t.name;
			} else {
				return GetPathToRoot(t.parent, root) + "/" + t.name;
			}
		}

		private IEnumerator AddControllers(){
			if(loaded) {
				SuperController.LogMessage("Already loaded. Reload plugin if needed.");
				yield break;
				
			}

			if(selectedRootPath == null){
				SuperController.LogMessage("No root node selected.");
				yield break;
			}

			Transform root = null;
			for(int i = 0; root == null && i < 100; i++){
				root = cuaRoot.Find(selectedRootPath);
				yield return null;
			}

			if(root == null){
				SuperController.LogMessage($"Could not find root node at path {selectedRootPath}");
				yield break;
			}

			
			// Wait a bit to let unity delete old objects in the case of reload
			yield return new WaitForSeconds(0.1f);


			// We need a person to get a joint control UI
			var personPrefab = SuperController.singleton.GetCachedPrefab("a_per", "Person");
			if(personPrefab == null){
				// No people in the scene yet. Load one (without adding to scene) and register it as the prefab. 
				// This might have unintended consequences. 
				SuperController.LogMessage("Person prefab not cached. Loading from asset bundle.");
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync("a_per", "Person", typeof(GameObject));
				if (request == null)
				{
					SuperController.LogError("Unable to load Person prefab");
					yield break;
				}
				yield return base.StartCoroutine(request);

				personPrefab = request.GetAsset<GameObject>();
				if (personPrefab != null)
				{
					SuperController.singleton.RegisterPrefab("a_per", "Person", personPrefab);
				} else {
					SuperController.LogError("Unable to load Person prefab");
					yield break;
				}
			}

			// Get new copy of the 'empty' prefab.
			AssetBundleLoadAssetOperation loadEmptyRequest = AssetBundleManager.LoadAssetAsync("a_mis", "Empty", typeof(GameObject));
			if (loadEmptyRequest == null)
			{
				SuperController.LogError("Unable to load Empty prefab");
				yield break;
			}
			yield return base.StartCoroutine(loadEmptyRequest);

			GameObject emptyPrefab = loadEmptyRequest.GetAsset<GameObject>();
			if (emptyPrefab == null) {
				SuperController.LogError("Unable to load Empty prefab");
				yield break;
			}

			
			
			if(controllerUIPrefab == null || sampleController == null) {
				GameObject sampleHipControl = null;
				foreach (GameObject go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]){
					if(go.name.Contains("UIFreeControllerJoint")) {
						controllerUIPrefab = go;
						continue;
					}
				}
			}

				//SuperController.LogMessage($"{sampleController.onPositionMesh}");
				
			if(controllerUIPrefab == null || sampleController == null) {
				SuperController.LogError("Unable to load required templates. Fix this by adding a Person to the scene");
				yield break;
			}

			// Initialize base objects
			
			
			// The top-level Atom is already initialized and has some internal state baked in.
			// We need to add the controllers to a new atom to make selection and timeline work
			string controlAtomUid = $"{containingAtom.uid}#Controllers";

			var existingControlAtom = SuperController.singleton.GetAtomByUid(controlAtomUid);

			// Remove the existing control atom
			if(existingControlAtom != null) {
				SuperController.singleton.RemoveAtom(existingControlAtom);
				while(existingControlAtom != null) {
					yield return null;
				}
			}

			emptyPrefab.SetActive(false);
			GameObject instantiatedEmpty = UnityEngine.Object.Instantiate<GameObject>(emptyPrefab);
			emptyPrefab.SetActive(true);
			controlAtom = instantiatedEmpty.GetComponent<Atom>();

			unityAssetControls = CreateChildObject(controlAtom.reParentObject.gameObject, "CustomUnityAssetControls");
			//controlUI = controlAtom.transform.Find("UIPlaceHolder").gameObject;
			controlUI = CreateChildObject(unityAssetControls, "UI");
			controllers = CreateChildObject(unityAssetControls, "Controllers");

			// Delete the dummy empty
			//SuperController.singleton.RemoveAtom(dummyEmpty);
			//Destroy(dummyEmpty);

			// SuperController.LogMessage($"Armature: {root}");
			ProcessRigging(root,0);

			if(createJoints.val){
				ConfigurableJoint jointToRootControl = CreateComponent(root.gameObject, typeof(ConfigurableJoint)) as ConfigurableJoint;
				jointToRootControl.connectedBody = containingAtom.mainController.GetComponent<Rigidbody>();

				jointToRootControl.xMotion = ConfigurableJointMotion.Locked;
				jointToRootControl.yMotion = ConfigurableJointMotion.Locked;
				jointToRootControl.zMotion = ConfigurableJointMotion.Locked;
				
				jointToRootControl.angularXMotion = ConfigurableJointMotion.Locked;
				jointToRootControl.angularYMotion = ConfigurableJointMotion.Locked;
				jointToRootControl.angularZMotion = ConfigurableJointMotion.Locked;
				JointDrive drive = new JointDrive();
				drive.positionSpring = 100000f;
				drive.positionDamper = 0f;
				drive.maximumForce = 10000f;
				jointToRootControl.xDrive = drive;
				jointToRootControl.yDrive = drive;
				jointToRootControl.zDrive = drive;
				jointToRootControl.angularXDrive = drive;
				jointToRootControl.angularYZDrive = drive;
			}
			
			
			try{
				foreach(Rigidbody rb in root.GetComponentsInChildren<Rigidbody>(true)){
					AddController(rb);
				}
				
				instantiatedEmpty.SetActive(true);

				Transform instantiatedControllers = SuperController.singleton.AddAtom(controlAtom, controlAtomUid, false, false, false, false);

			} catch (Exception e) {
				SuperController.LogMessage($"{e}");
			}

			foreach(var joint in controllerJoints){
				var rb = joint.connectedBody;
				joint.connectedBody = null;
				joint.connectedBody = rb;
			}

			controlAtom.parentAtom = containingAtom;
			
			yield return new WaitForSeconds(0.1f);


			RestoreControlAtom(savedJson);


			loaded = true;
			loadOnStartup.val = true;
			
			yield break;
		}
		
		private Rigidbody ProcessRigging(Transform parent, int depth){
			Rigidbody parentRB = parent.gameObject.GetComponent<Rigidbody>();
			if(parentRB == null && createRigidBodies.val){
				parentRB = CreateComponent(parent.gameObject, typeof(Rigidbody)) as Rigidbody;
				parentRB.isKinematic = false;
				parentRB.useGravity = false;
				parentRB.detectCollisions = true;
				parentRB.drag = defaultRbDrag;
				parentRB.mass = defaultRbMass;
				createdRigidbodies.Add(parentRB);
			}

			try{
			//	SuperController.LogMessage($"Processing: {parent}");
				for( int i = 0; i < parent.childCount; i++) {
					Transform child = parent.GetChild(i);
			//		SuperController.LogMessage($"Child: {child}");

					Rigidbody childRB = child.GetComponent<Rigidbody>();
					if(childRB == null){
						if(createRigidBodies.val){
					 		childRB = CreateComponent(child.gameObject, typeof(Rigidbody)) as Rigidbody;
							childRB.isKinematic = false;
							childRB.useGravity = false;
							childRB.detectCollisions = true;
							childRB.drag = defaultRbDrag;
							childRB.mass = defaultRbMass;
							createdRigidbodies.Add(childRB);
						} else {
							continue;
						}
					}
					
					if(parentRB !=null && createJoints.val){
							
						ConfigurableJoint childJoint = CreateComponent(child.gameObject, typeof(ConfigurableJoint)) as ConfigurableJoint;
						childJoint.connectedBody = parentRB;
						createdJoints.Add(childJoint);
						
						childJoint.xMotion = ConfigurableJointMotion.Locked;
						childJoint.yMotion = ConfigurableJointMotion.Locked;
						childJoint.zMotion = ConfigurableJointMotion.Locked;
						
						childJoint.angularXMotion = ConfigurableJointMotion.Limited;
						childJoint.angularYMotion = ConfigurableJointMotion.Limited;
						childJoint.angularZMotion = ConfigurableJointMotion.Limited;

						childJoint.rotationDriveMode = RotationDriveMode.Slerp;
						JointDrive slerpDrive = new JointDrive();
						slerpDrive.positionSpring = defaultJointDriveSpring;
						slerpDrive.positionDamper = defaultJointDriveDamper;
						slerpDrive.maximumForce = defaultJointDriveMaxForce;
						childJoint.slerpDrive = slerpDrive;
						
						SoftJointLimit xLowJointLimit = new SoftJointLimit();
						xLowJointLimit.limit = -180f;
						SoftJointLimit jointLimit = new SoftJointLimit();
						jointLimit.limit = 180f;
						childJoint.lowAngularXLimit = xLowJointLimit;
						childJoint.highAngularXLimit = jointLimit;
						childJoint.angularYLimit =jointLimit;
						childJoint.angularZLimit = jointLimit;
			
					}
					
					ProcessRigging(child, depth+1);
				}
				
			}
			catch (Exception ex)
			{
				SuperController.LogError($"Failed to init UnityObjectController: {ex}");
			}
			return parentRB;
		}

		private void AddController(Rigidbody rb){
			try {
				GameObject rbControl = CreateChildObject(controllers, $"{rb.gameObject.name}Control");
				rbControl.SetActive(false);
				rbControl.layer = 8;
				
				Rigidbody controlRB = CreateComponent(rbControl, typeof(Rigidbody)) as Rigidbody;
				controlRB.isKinematic = true;
				controlRB.useGravity = false;
				controlRB.detectCollisions = true;

				ConfigurableJoint controlJoint = CreateComponent(rb.gameObject, typeof(ConfigurableJoint)) as ConfigurableJoint;
				controlJoint.connectedBody = controlRB;
				controlJoint.autoConfigureConnectedAnchor = false;
				controlJoint.anchor = Vector3.zero;
				controlJoint.connectedAnchor = Vector3.zero;

				if( !controllerJoints.Contains(controlJoint)){
					controllerJoints.Add(controlJoint);
				}
			
				FreeControllerV3 controller = CreateComponent(rbControl, typeof(FreeControllerV3)) as FreeControllerV3;
				controller.containingAtom = controlAtom;
				controller.followWhenOff = rb.transform;
				controller.useGravityOnRBWhenOff = false;
				controller.meshScale = 0.5f;
				controller.selectedScale = 0.1f;
				controller.enableSelectRoot = false;
				controller.enabled = true;
				controller.hidden = false;
				controller.startingPositionState = FreeControllerV3.PositionState.Off;
				controller.startingRotationState = FreeControllerV3.RotationState.Off;
				
				if( !createdControllers.Contains(controller)){
					createdControllers.Add(controller);
				}

				// Copy meshes from the sample controller
				controller.material = sampleController.material;
				if(controller.onPositionMesh == null){
					controller.onPositionMesh = sampleController.onPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.onPositionMesh, controller.transform);
					controller.offPositionMesh = sampleController.offPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.offPositionMesh, controller.transform);
					controller.followingPositionMesh = sampleController.followingPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.followingPositionMesh, controller.transform);
					controller.holdPositionMesh = sampleController.holdPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.holdPositionMesh, controller.transform);
					controller.lockPositionMesh = sampleController.lockPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lockPositionMesh, controller.transform);
					controller.onRotationMesh = sampleController.onRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.onRotationMesh, controller.transform);
					controller.offRotationMesh = sampleController.offRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.offRotationMesh, controller.transform);
					controller.followingRotationMesh = sampleController.followingRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.followingRotationMesh, controller.transform);
					controller.holdRotationMesh = sampleController.holdRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.holdRotationMesh, controller.transform);
					controller.lockRotationMesh = sampleController.lockRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lockRotationMesh, controller.transform);
					controller.lookAtRotationMesh = sampleController.lookAtRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lookAtRotationMesh, controller.transform);
					controller.moveModeOverlayMesh = sampleController.moveModeOverlayMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.moveModeOverlayMesh, controller.transform);
					controller.rotateModeOverlayMesh = sampleController.rotateModeOverlayMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.rotateModeOverlayMesh, controller.transform);
					controller.deselectedMesh = sampleController.deselectedMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.deselectedMesh, controller.transform);
				}
				
				SphereCollider collider = CreateComponent(rbControl, typeof(SphereCollider)) as SphereCollider;
				collider.enabled = true;
				collider.radius = 0.1f;

				// Controller Link. Not sure what this is used for.
				GameObject controlLink = CreateChildObject(rbControl, rb.name+"Link");
				controlLink.layer = 8;
				
				SphereCollider linkCollider = CreateComponent(controlLink, typeof(SphereCollider)) as SphereCollider;
				linkCollider.enabled = true;
				linkCollider.radius = 0.01f;
				
				Rigidbody linkRB = CreateComponent(controlLink, typeof(Rigidbody)) as Rigidbody;
				linkRB.isKinematic = true;
				linkRB.useGravity = false;
				linkRB.detectCollisions = false;

				FreeControllerV3Link freeControllerLink = CreateComponent(controlLink, typeof(FreeControllerV3Link)) as FreeControllerV3Link;
				freeControllerLink.linkedController = controller;
				

				// UI
				GameObject uiPlaceholder = CreateChildObject(controlUI, $"{rb.gameObject.name}UIPlaceholder" );
				uiPlaceholder.SetActive(false);

				MoveAndRotateAsHUDAnchor moveAsHud = CreateComponent(uiPlaceholder, typeof(MoveAndRotateAsHUDAnchor)) as MoveAndRotateAsHUDAnchor;
				moveAsHud.scaleAsAnchor = true;
				moveAsHud.anchorNum = HUDAnchor.AnchorNum.Two;
									
				UIConnectorMaster connectorMaster = CreateComponent(uiPlaceholder, typeof(UIConnectorMaster)) as UIConnectorMaster;
				connectorMaster.UIPrefab = controllerUIPrefab.transform;
									
				UIConnector connector = CreateComponent(uiPlaceholder, typeof(UIConnector)) as UIConnector;
				connector.receiver = controller;
				connector.receiverTransform = rbControl.transform;
				
				controller.UITransforms = new Transform[]{uiPlaceholder.transform};
				controller.UITransformsPlayMode = new Transform[]{};
				
				rbControl.SetActive(true);
			} catch (Exception e) {
				SuperController.LogMessage($"{e}");
			}
		}

		public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = true, bool rightSide = false){
			JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
			storableFloat.storeType = JSONStorableParam.StoreType.Full;
			UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
			uiSlider.rangeAdjustEnabled = rangeAdjust;
			uiSlider.valueFormat = "F3";
			RegisterFloat(storableFloat);
			
			if(onChange != null){
				storableFloat.setCallbackFunction = onChange;
			}
			
			return storableFloat;
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false) {
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);

			JSONArray jsonArray = new JSONArray();
			controlAtom.Store(jsonArray,true,true);
			
			if(jsonArray.Count > 0) {
			//	SuperController.LogMessage($"storing {jsonArray[0]}");
				jc["controllerAtom"] = jsonArray[0];
			}
			jc["mainController"] = containingAtom.mainController.GetJSON();
			return jc;
		}

		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true){
            base.RestoreFromJSON(jc,restorePhysical,restoreAppearance,presetAtoms,false);
			savedJson = jc;
        }

		
        private void LoadPreset(string path) {
			if (string.IsNullOrEmpty(path))
				return;
			
			var presetJson = (JSONClass)LoadJSON(path);
			
			RestoreFromJSON(presetJson, true, true, null, false);

			if(loaded) {
				RestoreControlAtom(presetJson);
			} else {
				StartCoroutine("AddControllers");
			}
			
		}
		
		private void SavePreset(string path) {
			SuperController.singleton.fileBrowserUI.fileFormat = null;
			if (string.IsNullOrEmpty(path)) {
				return;
			}

			if (!path.ToLower().EndsWith($".{fileExt}")) {
				path += $".{fileExt}";
			}

			SaveJSON(GetJSON(true,true,true), path);
		}

		private void RestoreControlAtom(JSONClass jc) {
			if(jc == null) {
				return;
			}

			if(jc["mainController"] != null) {
				containingAtom.mainController.RestoreFromJSON(jc["mainController"] as JSONClass);
				var preset = containingAtom.GetStorableByID("Preset");
				preset.transform.position = containingAtom.mainController.transform.position;
				preset.transform.localPosition = containingAtom.mainController.transform.localPosition;
				preset.transform.rotation = containingAtom.mainController.transform.rotation;
			}

			JSONClass controllerAtomJson = jc["controllerAtom"] as JSONClass;
			if(controllerAtomJson != null && controlAtom != null) {
				controlAtom.Restore(controllerAtomJson);
				controlAtom.LateRestore(controllerAtomJson);
			}
		}
		
		private GameObject FindGameObject(GameObject parent, string name){
			Transform childTransform = parent.transform.Find(name);
			
			if(childTransform == null){
				return null;
			} else {
				return childTransform.gameObject;
			}
		}
		
		private GameObject CreateChildObject(GameObject parent, string name) {
			GameObject gameObject = new GameObject(name);
			gameObject.transform.parent = parent.transform;
			createdObjects.Add(gameObject);

			return gameObject;
		}
		
		private Component CreateComponent(GameObject parent, System.Type type) {
			Component component = parent.AddComponent(type) as Component;
			createdObjects.Add(component);
			return component;
		}

		public void OnDestroy()
		{
			foreach(FreeControllerV3 controller in createdControllers) {
				SuperController.singleton.GetAllFreeControllers().Remove(controller);
			}

			foreach(var o in createdObjects){
				Destroy(o);
			}
			
			Destroy(unityAssetControls);

			if(controlAtom != null){
				SuperController.singleton.RemoveAtom(controlAtom);
			}
		}
	}																
}